// Web++ Client

import java.net.*;
import java.io.*;
import java.util.*;
import java.applet.*;
import java.awt.*;


///////////////////////////////////////////////////////////////////////////////
// 
//
///////////////////////////////////////////////////////////////////////////////
public final class Webpp extends Applet
{

    // package data members ===================================================
    static LatTable lat;
    static Cache cache;				// Mehmet
    static RespTimeTable respTimes;		// Mehmet
    static String baseDir;
    static String latFilename;
    static String cacheFileName;		// Mehmet
    static String respTimeFileName;		// Mehmet
    static boolean useCacheIndex = true;	// Mehmet
    static boolean ClientHasParentDoc;	// Mehmet: set to true if parent HTML doc is in client cache
    static int os;
    static int browser;
    static boolean trace = false;
    static boolean timing = true;
    static boolean dummyRequestsStarted = false;	// Mehmet: used when determining the closest (host) server
    static boolean dummyRequestsFinished = false;	// Mehmet: used when determining the closest (host) server
    static String hostServer = null;			// Mehmet: chosen host server URL including its port number
    static String serverNames = null;		// Mehmet: names and port numbers of all servers
    static String dummyFileName = null;		// Mehmet: used when returned dynamic results to user (e.g. when choosing host server)
    static long responseTime = 0;	// Mehmet: stores the response time of current request, if successful
    static String prevResponseTime = null;	// Mehmet: stores the URL and response time of previous request
    
    // package constants ======================================================
    final static int TTL           = 900000; // TTL for stat entries (msec)
    final static long FAIL_PERIOD_INI = 60000; // initial fail period (msec)
    final static int STAT_SIZE     =  5;  // statistics array size
    final static double PERCENTILE = 55;  // percentile for server selection
    final static boolean ASSERT_ON = true;
    final static int UNIX          = 21;
    final static int WIN           = 22;
    final static int EXPLORER      = 23;
    final static int NAVIGATOR     = 24;
    final static int BUF_LEN       = 8192; // size of a communications buffer
    final static String NAME_SERVER = "begonia.ece.nwu.edu:8080";	// Name server (coordinator)
    
    // debug flags ============================================================
    // prints the selected server, dumps latency table and informs about asynch
    // samples
    final static boolean SERVER_CHOICE_DEBUG = false;
    
    // contains body of main for debugging
    final static boolean MAIN_DEBUG = false;

    // applet constructor =====================================================
    public void init()
    {
        String osName;
        String par;

	//System.out.println("Mehmet: in Webpp init");
        
        try {
            par = getParameter( "trace" );
            if( par != null && par.indexOf( "true" ) != -1 ){
                trace = true;
            } else {
                trace = false;
            }
            
            par = getParameter( "timing" );
            if( par != null && par.indexOf( "true" ) != -1 ){
                timing = true;
            } else {
                timing = false;
            }
            
            osName = System.getProperty( "os.name" );
            if( osName.indexOf( "indow" ) != -1 ){
                os = WIN;
                baseDir = new String( "C:\\TEMP\\" ); 
                if( trace ){
                    System.out.println( "OS: Windows" );
                }
            } else {
                os = UNIX;
                baseDir = new String( "/tmp/" );
                if( trace ){
                    System.out.println( "OS: Unix" );
                }
            }
            
            par = getParameter( "env" );
            Assert.assert( par != null );
            if( par.indexOf( "icrosoft" ) != -1 ){
                browser = EXPLORER;
                if( trace ){
                    System.out.println( "Browser: Explorer" );
                }
            } else {
                browser = NAVIGATOR;
                if( trace ){
                    System.out.println( "Browser: Navigator" );
                }
            }
           
            latFilename = new String( baseDir + "LatTable.ser" );
            File latFD = new File( latFilename );

	    // begin Mehmet
	    cacheFileName = new String( baseDir + "Cache.ser");
	    File cacheFD = new File( cacheFileName );
	    respTimeFileName = new String( baseDir + "RespTimes.ser");
	    File respTimeFD = new File( respTimeFileName );
	    // end Mehmet

            // Netscape security, will need to test for browser type once
            // we run on MS Explorer - should get dialog boxes out of
            // the way
            if( browser == NAVIGATOR ){
                //netscape.security.PrivilegeManager.enablePrivilege( 
                //                                     "UniversalFileRead" );
                //netscape.security.PrivilegeManager.enablePrivilege( 
                //                                     "UniversalFileWrite" );
                //netscape.security.PrivilegeManager.enablePrivilege( 
                //                                     "UniversalConnect" );
		netscape.security.PrivilegeManager.enablePrivilege(
							"Netcaster");
		//netscape.security.PrivilegeManager.enablePrivilege(
		//					"UniversalFileAccess");
		//netscape.security.PrivilegeManager.enablePrivilege(
		//					"Account Setup");
            } else {
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.FILEIO );
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.NETIO );
            }

            if( !latFD.exists() ){
                // create a new lat table
                lat = new LatTable();
            } else {
                // deserialize an existing lat table
                FileInputStream in = new FileInputStream( latFD );
                ObjectInputStream oin = new ObjectInputStream( in );
                lat = (LatTable) oin.readObject();
		oin.close();
            }

	    // begin Mehmet
	    if( !cacheFD.exists() ){
		// create a new cache index
		cache = new Cache();
	    } else {
		// deserialize an existing cache index
		FileInputStream in = new FileInputStream( cacheFD );
		ObjectInputStream oin = new ObjectInputStream( in );
		cache = (Cache) oin.readObject();
		oin.close();
		if (cache == null) {
		  cache = new Cache();
		}
	    }
	    if( !respTimeFD.exists() ){
		// create a new response time table
		respTimes = new RespTimeTable();
	    } else {
		// deserialize an existing response time table
		FileInputStream in = new FileInputStream( respTimeFD );
		ObjectInputStream oin = new ObjectInputStream( in );
		respTimes = (RespTimeTable) oin.readObject();
		oin.close();
		if (respTimes == null) {
		  respTimes = new RespTimeTable();
		}
		else hostServer = respTimes.getHostName();
	    }
	    // end Mehmet
        }

        catch( ClassNotFoundException e ){
            // create a new lat table
            lat = new LatTable();
	    cache = new Cache();		// Mehmet
	    respTimes = new RespTimeTable();	// Mehmet

            System.out.println( "init: lat table read exception: " +
                                e.getMessage() );
        }

        catch( IOException e ){
            // create a new lat table
            lat = new LatTable();
	    cache = new Cache();		// Mehmet
	    respTimes = new RespTimeTable();	// Mehmet

            if( Webpp.SERVER_CHOICE_DEBUG ){
                System.out.println( "init: lat table exception: " +
                                    e.getMessage() );
            }
        }

        catch( Exception e ){
            System.out.println( "init exception: " + e.getMessage() );
        }	
	}

    // applet destructor ======================================================
    public void destroy()
    {
	//System.out.println("In Webpp destructor");
        serLatTable();
	serCache();		// Mehmet: added this fcn call
	serRespTimesTable();	// Mehmet: added this fcn call
    }
    
    // Select a URL to send a request =========================================
    public String getUrl( String cArgs )
         throws Exception
    {   
        Timer tApplet = null;     // measures time to execute the applet
        Timer tResource = null;   // measures time to receive resource from net
        Vector args, embRes;	  // Mehmet: added embRes declaration
	String embStr;		  // Mehmet: string to add to request for embedded resources
        Hashtable servers;
        String bestUrl = null;
        String bestServer = null;
        Stat1 stat = null;
        boolean first = true;
	        
        if( Webpp.timing ){
            tApplet = new Timer();
            tApplet.start();
            tResource = new Timer();
            tResource.start();
        }

        // must re-enable priviledges to use them in this method
        if( browser == NAVIGATOR ){
            //netscape.security.PrivilegeManager.enablePrivilege( 
            //                                          "UniversalFileRead" );
            //netscape.security.PrivilegeManager.enablePrivilege( 
            //                                          "UniversalFileWrite" );
            //netscape.security.PrivilegeManager.enablePrivilege( 
            //                                          "UniversalConnect" );
	    netscape.security.PrivilegeManager.enablePrivilege(
							"Netcaster");
	    //netscape.security.PrivilegeManager.enablePrivilege(
		//					"UniversalFileAccess");
	    //netscape.security.PrivilegeManager.enablePrivilege(
		//					"Account Setup");
        } else {
            com.ms.security.PolicyEngine.assertPermission( 
                                  com.ms.security.PermissionID.FILEIO );
            com.ms.security.PolicyEngine.assertPermission( 
                                  com.ms.security.PermissionID.NETIO );
        }

	if (cArgs.indexOf("ServerList") >=0) {
	  dummyFileName = baseDir + "dummy.html";
	  //if (hostServer == null) findHostServer(null, this);	// Mehmet: determine the host server
	  findHostServer(null, this);		// Mehmet: determine the host server
	  return Net.patchPath(dummyFileName);
	}
        
	//System.out.println("Mehmet: cArgs=" + cArgs);	// Mehmet: debug
        args = getArgs( cArgs );
	servers = getServers( args );
        // Mehmet: added the following two lines to extract the list
        // of embedded resources from the URL (arguments of applet),
 	// and to check which ones are already in client cache.
        embRes = getEmbeddedResources( args );
	embStr = cache.checkEmbResources(embRes);
	//System.out.println("Mehmet: embStr=" + embStr);	// Mehmet: debug

        while( (bestServer = lat.getBestServer(servers)) != null ){
	    //System.out.println("Mehmet: bestServer: " + bestServer);	// Mehmet: debug
            Assert.assert( bestServer != null );

            if( Webpp.SERVER_CHOICE_DEBUG ){
                System.out.println( "best server: " + bestServer );
                System.out.println( "lat. table: " );
                lat.print();
            }
            
            if( Webpp.trace ){
                System.out.println( "selected server: " + bestServer );
            }
            
            bestUrl = (String) servers.get( bestServer );
	    //System.out.println("Mehmet: bestUrl: " + bestUrl);	// Mehmet: debug
            Assert.assert( bestUrl != null );

            if( first && Webpp.timing ){
                tApplet.stop();
                System.out.println( "time for getUrl() exec = " + 
                                    tApplet.getTime() );
                first = false;
            }

	    // Begin Mehmet: get previous resp time, if there is one
	    //	Important note: this has to be done before any call to Net.getDoc() since it is used there.
	    if (respTimes.containsKey(bestServer))
	      prevResponseTime = (String) respTimes.get(bestServer);
	    else
	      prevResponseTime = null;
	    // End Mehmet

            showStatus( "Contacting " + bestServer + "..." );
	    // Mehmet: added if-else statements here
	    if (embStr != null)	
                stat = Net.getDoc( bestUrl + embStr, this); 
	    else
	        stat = Net.getDoc( bestUrl, this );
            
            if( Webpp.timing ){
                tResource.stop();
                System.out.println( "time to get resource = " + 
                                    tResource.getTime() );
		responseTime = tResource.getTime();	// Mehmet: keep response time
            }

            if( stat != null ){
	        // successfully retrieved the document
                showStatus( "Contacted " + bestServer );
	        
	        if( Webpp.SERVER_CHOICE_DEBUG ){
                    System.out.println( "rec OK, latency = " + stat.lat );
                }
            
                lat.addLatency( bestServer, stat.lat );

		// Begin Mehmet: insert response time to the respTimes table
		if (responseTime > 0) {
		  String fnameStr = new String (bestUrl);
		  // Take the only filename portion of the URL if possible (i.e. unless URL ends with '/')
		  int separatorPos = bestUrl.lastIndexOf('/');
	  	  if (separatorPos != -1 && separatorPos < (bestUrl.length() - 1))
		    fnameStr = bestUrl.substring(separatorPos + 1);

		  // Mehmet: use the next line if you want to piggyback response time.
		  //	Otherwise, use the line following it in order to piggyback the latency.
		  //String rtime = (new Long(responseTime)).toString();
		  String rtime = (new Long((long) (stat.lat * 1000))).toString();

		  respTimes.put(bestServer, fnameStr + "|" + rtime);
		}
		// End Mehmet
                break;
                
            } else {
                // mark server as unreachable
                showStatus( "Could not reach " + bestServer );
                
                lat.updateReachability( bestServer, false );
                
                if( Webpp.trace ){
                    System.out.println( "server " + bestServer + 
                                        " unreachable" );
                }
            }
	}

        // asynchronously write lat table to disk
        // guarantees no lost updates for next invokation of the applet
        // (as there is no guarantee that destroy() gets called beforehand)
        chptLatTable( this );
	
	chptCache( this );	// Mehmet: asynchronously write cache index to disk

	chptRespTimes( this );	// Mehmet: asynchronously write respTimes table to disk
        
        if( stat != null ){
            return stat.locURL;
        } else {
            // failure to find the requested document should be 
            // handled more gracefully
            return null;
        }
    }
    
    
    // serialize lat table onto disk =======================================
    void serLatTable()
    {
        FileOutputStream out;
        ObjectOutputStream oout;
        
        // serialize current latency table
        Assert.assert( latFilename != null );

        try {
            if( browser == NAVIGATOR ){
                netscape.security.PrivilegeManager.enablePrivilege( 
                                                      "UniversalFileRead" );
                netscape.security.PrivilegeManager.enablePrivilege( 
                                                      "UniversalFileWrite" );
            } else {
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.FILEIO );
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.NETIO );
            }
            
            out  = new FileOutputStream( latFilename );
            oout = new ObjectOutputStream( out );

            oout.writeObject( lat );
            oout.flush();
            oout.close();
        }

        catch( IOException e ){
            System.out.println( "destroy exception: " + e.getMessage() );
        }
    }


    // Mehmet: serialize cache onto disk =======================================
    void serCache()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;
        
        // serialize current cache index
        Assert.assert( cacheFileName != null );

        try {
            if( browser == NAVIGATOR ){
                netscape.security.PrivilegeManager.enablePrivilege( 
                                                      "UniversalFileRead" );
                netscape.security.PrivilegeManager.enablePrivilege( 
                                                      "UniversalFileWrite" );
            } else {
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.FILEIO );
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.NETIO );
            }
            
	    if (cache != null && cache.size() > 0) {
              out  = new FileOutputStream( cacheFileName );
	      if (out != null)
                oout = new ObjectOutputStream( out );
	    }

	    if (oout != null) {
              oout.writeObject( cache );
              oout.flush();
              oout.close();
	    }
        }

        catch( IOException e ){
            System.out.println( "serCache exception: " + e.getMessage() );
        }
    }


    // Mehmet: serialize respTimes table onto disk =======================================
    void serRespTimesTable()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;
        
        // serialize current response times table
        Assert.assert( respTimeFileName != null );

        try {
            if( browser == NAVIGATOR ){
                netscape.security.PrivilegeManager.enablePrivilege( 
                                                      "UniversalFileRead" );
                netscape.security.PrivilegeManager.enablePrivilege( 
                                                      "UniversalFileWrite" );
            } else {
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.FILEIO );
                com.ms.security.PolicyEngine.assertPermission( 
                                      com.ms.security.PermissionID.NETIO );
            }
            
	    if (respTimes != null && respTimes.size() > 0) {
              out  = new FileOutputStream( respTimeFileName );
	      if (out != null)
                oout = new ObjectOutputStream( out );
	    }

	    if (oout != null) {
              oout.writeObject( respTimes );
              oout.flush();
              oout.close();
	    }
        }

        catch( IOException e ){
            System.out.println( "serRespTimesTable exception: " + e.getMessage() );
        }
    }


    // asynchronously serialize latency table to disk =======================
    private void chptLatTable( Webpp self )
    {
        new ChptLatTable( self ).run();
    }


    // Mehmet: asynchronously serialize cache index to disk =======================
    private void chptCache( Webpp self )
    {
        new ChptCache( self ).run();
    }


    // Mehmet: asynchronously serialize respTimes table to disk =======================
    private void chptRespTimes( Webpp self )
    {
        new ChptRespTimes( self ).run();
    }


    // Mehmet: find the closest (host) server to this client ======================
    public void findHostServer(String serverNames, Webpp self)
    {
	dummyRequestsStarted = true;
	dummyRequestsFinished = false;
	//new HostServerFinder( self, serverNames ).run();
	HostServerFinder hsf = new HostServerFinder( self, serverNames );
	hsf.runit();
    }
    
    
    // Unpack arguments from a string =======================================
    private Vector getArgs( String cArgs )
    {
        StringTokenizer st = new StringTokenizer( cArgs, " " );
        Vector argv = new Vector();

        for (; st.hasMoreTokens(); argv.addElement( st.nextToken() ) );

        return argv;        
    }

    // Get the embedded resources list out of the args vector
    private Vector getEmbeddedResources( Vector args )
    {
      Enumeration e = args.elements();
      String embRes = (String) e.nextElement();
      StringTokenizer st = new StringTokenizer(embRes, ";");
      Vector embResVec = new Vector();
      for (; st.hasMoreTokens(); embResVec.addElement( st.nextToken() ) );

      return embResVec;
    }

    // Filter out a list of servers from a list of URLs =======================
    //    (the hashtable maps each server to the corresponding URL)
    private Hashtable getServers( Vector urls )
    {
        Hashtable servers = new Hashtable();

        try {
            String urlS;
            URL    url;
            String host;
            int    port;
            String server;

            Enumeration e = urls.elements();
            urlS = (String) e.nextElement();
            for ( ; e.hasMoreElements(); ) {

                urlS = (String) e.nextElement();
		//System.out.println("Mehmet: next element=" + urlS);	// Mehmet: debug
                url  = new URL( urlS );
                host = url.getHost();
                port = url.getPort();

                if( port == -1 ){
                    server = host;
                } else {
                    server = host + ":" + Integer.toString( port );
                }

                servers.put( server, urlS );
            }
        }

        catch( MalformedURLException e1 ){
	        System.out.println( e1.getMessage() );
        }

        return servers;
    }

    // Get fastest server =====================================================
    private String getBestUrl( Hashtable servers )
        throws Exception
    {
        return lat.getBestServer( servers );
    }
    
    // write a string to Java console (used for instrumenting HTML) ===========
    public void log( String msg )
    {
        System.out.println( msg );
    }
    
    // main ===================================================================
    public void main( String[] argv )		// Webpp class
    {
        String url;

        if ( MAIN_DEBUG ) {
            try {           
	            for (int i = 0; i < 2; i++) {
                    System.out.println( "iteration " + i );
                
                    url = getUrl( "http://coke.stanford.edu:8088/w05K.txt" +
                         " " + "http://frankfurt.cs.uni-sb.de:8088/w05K.txt" );

                    Thread.currentThread().sleep( 4000 );
                }            
            }
            
            catch ( Exception e ) {
                System.out.println( "Exception in main: " + e.getMessage() );
                Runtime.getRuntime().exit( -1 );
            }                     
        }   
    }
}

///////////////////////////////////////////////////////////////////////////////
// 
// Thread for asynchronous lat table checkpointing
//
///////////////////////////////////////////////////////////////////////////////
final class ChptLatTable extends Thread
{
    // constructor ============================================================
    ChptLatTable( Webpp app )
    {
        // remember reference to parent applet
        applet_ = app;
    }
    
    // asynchronously writes lat table to disk =============================
    public void run()
    {          
        applet_.serLatTable();
    }

    // private data members ===================================================
    private Webpp applet_;
}


///////////////////////////////////////////////////////////////////////////////
// 
// Mehmet: Thread for asynchronous cache index checkpointing
//
///////////////////////////////////////////////////////////////////////////////
final class ChptCache extends Thread
{
    // constructor ============================================================
    ChptCache( Webpp app )
    {
        // remember reference to parent applet
        applet_ = app;
    }
    
    // asynchronously writes cache index to disk =============================
    public void run()
    {          
        applet_.serCache();
    }

    // private data members ===================================================
    private Webpp applet_;
}


///////////////////////////////////////////////////////////////////////////////
// 
// Mehmet: Thread for asynchronous response times table serialization
//
///////////////////////////////////////////////////////////////////////////////
final class ChptRespTimes extends Thread
{
    // constructor ============================================================
    ChptRespTimes( Webpp app )
    {
        // remember reference to parent applet
        applet_ = app;
    }
    
    // asynchronously writes response times table to disk =============================
    public void run()
    {          
        applet_.serRespTimesTable();
    }

    // private data members ===================================================
    private Webpp applet_;
}


///////////////////////////////////////////////////////////////////////////////
// 
// Mehmet: Thread for determining the closest (host) server
//
///////////////////////////////////////////////////////////////////////////////

final class HostServerFinder extends Thread
{
    // constructor ============================================================
    HostServerFinder( Webpp app, String serverNames )
    {
        // remember reference to parent applet
        applet_ = app;

	// get serverNames string from the argument
	this.serverNames = serverNames;
    }


    // Mehmet: getDummyURL() ================================================
    //	performs a get request to the dummy.html at the
    //	given server to measure its latency. This function is called when
    //	the Web++ client receives the list of server names from a servlet
    //	and starts measuring their latencies to pick the closest server
    //	(host server) to itself.
    public long getDummyURL(String urlName) throws Exception {
      long first_byte_time, bytesRead;
      long time1, time2=0;
      byte b = 0;
      int done = 0;

      if( applet_.browser == applet_.NAVIGATOR ){
        //netscape.security.PrivilegeManager.enablePrivilege( 
        //                                          "UniversalFileRead" );
        //netscape.security.PrivilegeManager.enablePrivilege( 
        //                                          "UniversalFileWrite" );
        //netscape.security.PrivilegeManager.enablePrivilege( 
        //                                          "UniversalConnect" );
	netscape.security.PrivilegeManager.enablePrivilege(
							"Netcaster");
      } else {
        com.ms.security.PolicyEngine.assertPermission( 
                              com.ms.security.PermissionID.FILEIO );
        com.ms.security.PolicyEngine.assertPermission( 
                              com.ms.security.PermissionID.NETIO );
      }

      URL url = new URL("http://" + urlName + "/dummy.html");
      URLConnection conn = url.openConnection();
      //PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("out.txt")));
    
      bytesRead = 0;
      time1 = System.currentTimeMillis();
      DataInputStream in = new DataInputStream(conn.getInputStream());
      try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      while (done == 0) {
        // Mehmet: comment out the next three lines if you don't want to echo the downloaded file to screen
        //System.out.print((char)b);
        //if (b==13 || b==10) out.println("");
        //else out.print((char)b);
        bytesRead++;
        if (bytesRead == 1) { time2 = System.currentTimeMillis(); }
        try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      }
      //out.close();
      in.close();
    
      first_byte_time = time2 - time1;
      if (bytesRead > 0 && time2 > 0) return first_byte_time;
      else return 0;
    }


    // Mehmet: doPostRequest() =============================================
    //	performs a POST request for the given request type to the given server.
    public String doPostRequest(String urlName, String reqType) throws Exception {
      long bytesRead;
      byte b = 0;
      int done = 0;

      //yield();
      //System.out.println("Before Sleep: " + System.currentTimeMillis());
      //sleep(10000);
      //System.out.println("After  Sleep: " + System.currentTimeMillis());

      try {
      if( applet_.browser == applet_.NAVIGATOR ){
        //netscape.security.PrivilegeManager.enablePrivilege( 
        //                                          	"UniversalFileRead" );
        //netscape.security.PrivilegeManager.enablePrivilege( 
        //                                          	"UniversalFileWrite" );
        //netscape.security.PrivilegeManager.enablePrivilege( 
        //                                          	"UniversalConnect" );
	netscape.security.PrivilegeManager.enablePrivilege(
							"Netcaster");
	//netscape.security.PrivilegeManager.enablePrivilege(
	//						"UniversalFileAccess");
	//netscape.security.PrivilegeManager.enablePrivilege(
	//						"Account Setup");
      } else {
        com.ms.security.PolicyEngine.assertPermission( 
                              com.ms.security.PermissionID.FILEIO );
        com.ms.security.PolicyEngine.assertPermission( 
                              com.ms.security.PermissionID.NETIO );
      } 
      } catch (Exception e) {
	System.out.println("Error in doPostRequest setpriv.: " +
						e.getMessage() );
	return null;
      }

      String str = new String(urlName + "/servlet/webppserv.Webpp/?" + reqType);
      URL url = new URL(str);
      System.out.println("DoPost: " + url);
      //HttpURLConnection conn = null;
      URLConnection conn = null;
      try {
	//conn = (HttpURLConnection) url.openConnection();
	conn = url.openConnection();
      } catch (Exception e) {
	System.out.println("Error in doPostRequest openConnection: " +
						e.getMessage() );
	return null;
      }

//      conn.setDoOutput(true);
//      try {
//	  conn.setRequestMethod("POST");
//      } catch (ProtocolException pe) {
//	System.out.println("Error: cannot set reqType!");
//      }

      String result = null;
      StringBuffer buf = new StringBuffer();
      //PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("out.txt")));
      bytesRead = 0;
    
      //DataOutputStream outs = new DataOutputStream(conn.getOutputStream());
      //outs.writeUTF(reqType);
      //outs.close();

      DataInputStream in = new DataInputStream(conn.getInputStream());
      try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      while (done == 0) {
        // Mehmet: comment out the next three lines if you don't want to echo the downloaded file to screen
        //System.out.print((char)b);
        //if (b==13 || b==10) out.println("");
        //else out.print((char)b);
	buf.append((char) b);
        bytesRead++;
        try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      }
      //out.close();
      in.close();

      if (buf.length() > 0)
	result = buf.toString();
    
      if (bytesRead > 0) return result;
      else return null;
    }
    

    //public void run()
    public void runit()
    {          
	String bestServer = null, url = null;
	long bestLatency = 0;
	FileOutputStream fos = null;
	try {
	  fos = new FileOutputStream(applet_.dummyFileName);
	} catch (IOException ioe) {
	  System.out.println("Error: cannot create " + applet_.dummyFileName);
	  return;
	}
	PrintWriter dummyOut = new PrintWriter(fos);

	dummyOut.println("Choosing the host server...<BR>");
	
	//yield();

	// If serverNames string is empty, then request the serverNames list from
	// the name server (coordinator) by sending a POST request.
	if (serverNames == null) {
	  url = new String("http://" + applet_.NAME_SERVER);
	  try {
	    serverNames = doPostRequest(url, "ServerList");
	  } catch (Exception e) { serverNames = null; }
	  //System.out.println("Received ServerList is : " + serverNames);
	  dummyOut.println("Received ServerList is : " + serverNames + "<BR>");
	}

	if (serverNames == null) {
	  //System.out.println("Error: ServerList is empty in HostServerFinder::run()");
	  dummyOut.println("Error: cannot get the server list!<BR>");
	  dummyOut.close();
	  return;
	}

      // First, send GET requests to all servers and measure latencies.
	// The function getDummyURL() is used for requesting the dummy.html
	// at each server.
	StringTokenizer st;
	long latency = 0;
	st = new StringTokenizer(serverNames, " ");
	while (st.hasMoreTokens()) {
	  url = (String) st.nextToken();
	  try {
	    latency = getDummyURL(url);
	  } catch (Exception e) { latency = 0; }
	  if (latency > 0) {
	    // The next line inserts the latency to server latency table so that 
	    //	it can be used for later bestURL selections
	    //applet_.lat.insertLatency( url, (new Long(latency)).doubleValue() / 1000.0);
	    //System.out.println("DummyRequest URL: " + url + "  Latency: " + latency);
	    dummyOut.println("URL: " + url + " Latency: " + latency + "<BR>");
	    if (bestLatency == 0 || latency < bestLatency) {
	      bestLatency = latency;
	      bestServer = url;
	    }
	  }
	  else {
	    //System.out.println("DummyRequest no reply from " + url);
	    dummyOut.println("No reply from " + url + "<BR>");
	  }
	}
	
	if (bestServer == null) {
	  System.out.println("Error in HostServerFinder::run() : cannot find a host server!");
	  dummyOut.println("Error: cannot find a host server!<BR>");
	  dummyOut.close();
	  //yield();
	  return;
	}

	// Next, set the static variable hostServer to the name of the closest
	// server in Webpp class.

	applet_.hostServer = bestServer;
	applet_.dummyRequestsFinished = true;
	applet_.dummyRequestsStarted = false;
	applet_.respTimes.putHostName(applet_.hostServer);
	System.out.println("Host server chosen: " + applet_.hostServer);
	dummyOut.println("Host server chosen: " + applet_.hostServer + "<BR>");

	//yield();

	// Finally, send a POST request to the closest server to let it know
	// about that.

	if (applet_.hostServer != null) {
	  try {
	    String result = doPostRequest("http://" + applet_.hostServer, "HostServer");
	  } catch (Exception e) {
	    System.out.println("Error: Cannot send HostServer msg to my host!");
	  }
	}

	dummyOut.close();	// close dummy file so that browser can display it
    }

    // private data members ===================================================
    private Webpp applet_;
    private String serverNames;
}