package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;



///////////////////////////////////////////////////////////////////////////////
//
// Trivial cache with no replacement
//
///////////////////////////////////////////////////////////////////////////////
class Cache extends Hashtable implements Serializable
{
    /* class variables */
    protected static int highInx = 0;
    protected static boolean created = false;
    protected static String serFileName = null;
    protected static File cacheDir;
    protected static String cacheDirName;
    
    /* instance variables */
    protected boolean closed = false;
    
    /* reference to parent servlet */
    protected Webpp servlet;
    
   /*
    *=========================================================================
    * Constructor
    *=========================================================================
    */    
    protected Cache()
    {
        /* intentionally left empty */
    }
    
   /*
    *=========================================================================
    * Cache factory
    *=========================================================================
    */
    public static Cache factory( Webpp parentServlet )
    {
        File fd = null;
        FileInputStream in = null;
        ObjectInputStream oin = null;
        Cache cache = null;
        boolean status;
        
        if( created ){
            // cache is singleton 
            // Assert.assert( false );
            System.out.println( Webpp.webppRootDir + 
                                ": detected another cache instance" );
        }
        
        cacheDirName = Webpp.webppRootDir + 
                       System.getProperty( "file.separator" ) +
                       "cache";
        cacheDir = new File( cacheDirName );
      
        if( !cacheDir.exists() ){
            status = cacheDir.mkdirs();
            Assert.assert( status == true, 
                           "Cache.factory: could not create " + 
                           cacheDirName );
        }
      
        serFileName = cacheDirName + System.getProperty( "file.separator" ) +
                      "cache.ser";
        fd = new File( serFileName );

        if( fd.exists() ){
            /* deserialize an existing cache */
            if( Webpp.debugCache ){
                System.out.println( "***: deserializing cache from " + 
                                    serFileName );
            }
            
            try {
                in = new FileInputStream( fd );
            } catch( Exception e ){
                Assert.assert( false, e.getMessage() );
            }
            
            try {
                oin = new ObjectInputStream( in );
            } catch( Exception e ){
                Assert.assert( false, e.getMessage() );
            }

            try {
                cache = (Cache) oin.readObject();
            } catch( Exception e ){
                Assert.assert( false, e.getMessage() );
            }

        } else {
            /* create a new cache */
            if( Webpp.debugCache ){
                System.out.println( "***: creating a new cache" );
            }
            
            cache = new Cache();
        }
        
        cache.servlet = parentServlet;
        
        created = true;
        return cache;
    }

   /*
    *=========================================================================
    * get a file either from the cache or from disk (or network)
    *=========================================================================
    */
    public BufferedInputStream getFile( String fileName ) 
        throws FileNotFoundException
    {
        BufferedInputStream in = null;
        String cacheFile = null;

	// Mehmet: Servlet cache was disabled when I received Webpp code.
	// I keep servlet cache disabled in order to resolve
	// problem of keeping track of which embedded resources
	// are in the cached file and which ones are already in
	// the client cache. It is beneficial to cache the whole
	// package with all embedded resources. Caching the partial
	// list of embedded resources with the requested page will
	// complicate things too much.

        if( Webpp.useCache && 
            (cacheFile = (String) get(fileName)) != null ){
            /* cache hit */
            if( Webpp.debugCache ){
                System.out.println( "***: file " + fileName + 
                                    " - cache hit " );
            }
            
            in = new BufferedInputStream( new FileInputStream(cacheFile),
                                              Webpp.BUFFER_SIZE );
            return in;

        } else {
            /* cache miss */
            if( Webpp.debugCache ){
                System.out.println( "***: file " + fileName + 
                                    " cache miss " );
            }
            
            cacheFile = cacheDirName + 
                        System.getProperty( "file.separator" ) +
                        "cache" + 
                        Integer.toString( highInx );
            highInx++;

            in = new SubstInputStream( fileName, cacheFile, servlet );
            put( fileName, cacheFile );

            return in;
        }
    }

   /*
    *=========================================================================
    * close cache by serializing cache object on the disk
    *=========================================================================
    */
    public void close()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;

        if( !Webpp.useCache ){
            return;
        }
        
        try {
            out = new FileOutputStream( serFileName );
        } catch( IOException e ){
            Assert.assert( false, e.getMessage() );
        }
        try {
            oout = new ObjectOutputStream( out );
        } catch( IOException e ){
            Assert.assert( false, e.getMessage() );
        }
        
        try {
            oout.writeObject( this );
        } catch( IOException e ){
            Assert.assert( false, e.getMessage() );
        } 
        try {
            oout.flush();
        } catch( IOException e ){
            Assert.assert( false, e.getMessage() );
        }
        try {
            oout.close();    
        } catch( IOException e ){
            Assert.assert( false, e.getMessage() );
        }
        
        closed = true;
    }
    
   /*
    *=========================================================================
    * serialized on the disk if not done
    *=========================================================================
    */
    protected void finalize()
    {
        if( !closed ){
            close();
        }
    }
}
