package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;




///////////////////////////////////////////////////////////////////////////////
//
// Trivial cache with no replacement
//
///////////////////////////////////////////////////////////////////////////////
class ReplicationDir extends Hashtable implements Serializable
{
    /* class variables */
    protected static boolean created = false;
    protected static String serFileName = null;
    protected static File replDirFd;
    protected static String replDirName;    
    protected static ReplicationDir self = null;
    
    /* instance variables */
    protected boolean closed = false;
    
   /*
    *=========================================================================
    * Constructor
    *=========================================================================
    */    
    protected ReplicationDir()
    {
        super();
        /* intentionally left empty */
    }
    
   /*
    *=========================================================================
    * Replication Directory factory
    *=========================================================================
    */
    public static ReplicationDir factory()
    {
        File fd = null;
        BufferedReader in = null;
        ReplicationDir replDir = null;
        boolean status;
        String line = null;
        String lurl;
        String purl;
        int pos;
        
        if( created ){
            // replication directory is singleton 
            // Assert.assert( false );
            System.out.println( Webpp.webppRootDir +
                                ": detected another instance of repl dir" );                                
        }
         
        replDirName = Webpp.webppRootDir + 
                      System.getProperty( "file.separator" ) +
                      "replication";
        replDirFd = new File( replDirName );
      
        if( !replDirFd.exists() ){
            status = replDirFd.mkdirs();
            Assert.assert( status == true, 
                           "ReplicationDir.factory: could not create " + 
                           replDirName );
        }
      
        serFileName = replDirName + System.getProperty( "file.separator" ) +
                      "replicationDir.ser";
        fd = new File( serFileName );

        replDir = new ReplicationDir();
        
        if( fd.exists() ){
            /* deserialize an existing replication directory */
            if( Webpp.debugReplDir ){
                System.out.println( "***: deserializing " + serFileName );
            }
            
            try {
                in = new BufferedReader( new FileReader( fd ) );
            } catch( Exception e ){
                Assert.assert( false, e.getMessage() );
            }
        
            try {
                while( (line = in.readLine()) != null ){
                    if( Webpp.debugReplDir ){
                        System.out.println( line );
                    }
                    
                    pos = line.indexOf( ' ' );
                    if( pos == -1 ){
                        continue;
                    }
                
                    lurl = line.substring( 0, pos );
                    purl = line.substring( pos+1, line.length() );
                
                    if( lurl.length() == 0 || purl.length() == 0 ){
                        continue;
                    }
                    
                    replDir.put( lurl, purl );
                }
            } catch( Exception e ){
                Assert.assert( false, e.getMessage() );
            }   
            
        } else {
            if( Webpp.debugReplDir ){
                System.out.println( "***: creating a new repl dir" );
            }
        }
        
        try {
            in.close();
        } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
        }
        
        created = true;
        self    = replDir;
        return replDir;
    }

   /*
    *=========================================================================
    * get list of URLs related to a given LURL
    *=========================================================================
    */
    public static String getUrls( String key )
    {
        String result;
        
        Assert.assert( key != null );
        
        result = (String) self.get( key );
        
        if( Webpp.debugReplDir ){
            System.out.println( "***: lookup result for " + key + ":: " +
                                result );
        }
        //return (String) self.get( key );
	return result;
    }

   /*
    *=========================================================================
    * append a URL to a list of replicas of a given LURL
    *=========================================================================
    */
    public static void appendUrl( String key, String url )
    {
        String urlList = null;
        
        Assert.assert( key != null && url != null, key + " " + url );
        
        urlList = (String) self.get( key );
        urlList += " " + url;
        self.put( key, urlList );
        
        if( Webpp.debugReplDir ){
            System.out.println( "***: appending " + url + 
                                " to replication string; new value " +
                                urlList );
        }
    }

   /*
    *=========================================================================
    * Mehmet: added the following function
    * get the list of embedded resources for a given LURL (logical URL)
    *=========================================================================
    */
    public static String getEmbeddedResources(String key)
    {
	int pos = key.lastIndexOf(System.getProperty("file.separator"));
	String tmp = null;
	if (pos != -1) tmp = key.substring(pos + 1);
	else tmp = key;
	tmp = getUrls(tmp);
	if (tmp == null) return null;
	String tmp2 = null;

	//StringTokenizer st = new StringTokenizer(tmp);
	//if (st.hasMoreElements())
	//  tmp2 = (String) st.nextElement();	// get emb resources string

	pos = tmp.indexOf(' ');
	if (pos != -1) tmp2 = tmp.substring(0, pos);

	if (tmp2.equalsIgnoreCase("NULL")) return null;
	else return tmp2;
    }

   /*
    *=========================================================================
    * Mehmet: added the following function
    * getLocalURL() : get the URL for the local replica of the given resource
    *=========================================================================
    */
    public static String getLocalURL(String key)
    {
	String tmp = getUrls(key);
	if (tmp == null) return null;	// get the replication dir record

	String tmp2 = null;
	StringTokenizer st = new StringTokenizer(tmp);
	if (st.hasMoreElements())
	  tmp2 = (String) st.nextElement();	// skip the embedded resources part
	while (st.hasMoreElements()) {
	  tmp2 = (String) st.nextElement();
	  if (tmp2.indexOf(Webpp.server) != -1) return tmp2;	// found the URL
	}
	return null;	// not found, return null
    }
 
   /*
    *=========================================================================
    * close replication directory by serializing it on the disk
    *=========================================================================
    */
    public void close()
    {
        BufferedWriter out = null;
        Enumeration enum;
        String line;
        String lurl;
        String purl;
        
        
        try {
            out = new BufferedWriter( new FileWriter( serFileName ) );
        } catch( IOException e ){
            Assert.assert( false, e.getMessage() );
        }
       
        if( Webpp.debugReplDir ){
            System.out.println( "***: serializing repl dir to " + 
                                serFileName );
        }
        
        enum = keys();
        try {
            while( enum.hasMoreElements() ){
                lurl = (String) enum.nextElement();
            
                purl = (String) get( lurl );
                Assert.assert( purl != null, "close: " + lurl );
            
                out.write( lurl + " " + purl );
                out.newLine();
                
                if( Webpp.debugReplDir ){
                    System.out.println( lurl + " " + purl );
                }
            }
        } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
        }
        
        try {
            out.close();
        } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
        }
        
        closed = true;
    }
    
   /*
    *=========================================================================
    * serialize on the disk if not done
    *=========================================================================
    */
    protected void finalize()
    {
        if( !closed ){
            close();
        }
    }
}   
