package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;




///////////////////////////////////////////////////////////////////////////////
//
// ResponseTime Table
//
///////////////////////////////////////////////////////////////////////////////
class ResponseTimeTable extends Hashtable implements Serializable
{
    /* class variables */
    protected static boolean created = false;
    protected static boolean updated = false;
    protected static String serFileName = null;
    protected static File cTableFd;
    protected static String cTableDirName;    
    protected static ResponseTimeTable self = null;
    protected static int noOfSegments;
    
    /* instance variables */
    protected boolean closed = false;
    
   /*
    *=========================================================================
    * Constructor
    *=========================================================================
    */    
    protected ResponseTimeTable()
    {
        super();
        /* intentionally left empty */
    }
    
   /*
    *=========================================================================
    * Replication Directory factory
    *=========================================================================
    */
    public static ResponseTimeTable factory(int noOfSegments)
    {
        File fd = null;
        ResponseTimeTable cTable = null;
        boolean status;
        String line = null;
        String thekey;
        String thevalue;
        int pos;
        
        if( created ){
            // response time table is singleton 
            // Assert.assert( false );
            System.out.println( Webpp.webppRootDir +
                                ": detected another instance of response time table" );                                
        }

/*
        cTableDirName = Webpp.webppRootDir + 
                      System.getProperty( "file.separator" ) +
                      "replication";
        cTableFd = new File( cTableDirName );
      
        if( !cTableFd.exists() ){
            status = cTableFd.mkdirs();
            Assert.assert( status == true, 
                           "ResponseTimeTable.factory: could not create " + 
                           cTableDirName );
        }
      
        serFileName = cTableDirName + System.getProperty( "file.separator" ) +
                      "responseTimeTable.ser";
        fd = new File( serFileName );

	if( !fd.exists() ){
            // create a new response time table
	      cTable = new ResponseTimeTable();
        } else {
            // deserialize an existing response time table
		try {
              FileInputStream in = new FileInputStream( fd );
              ObjectInputStream oin = null;
		  if (in != null) {
		    oin = new ObjectInputStream( in );
		  }
		  if (oin != null) {
                    cTable = (ResponseTimeTable) oin.readObject();
		    oin.close();
		  }
		  else {
		    System.out.println( "ResponseTimeTable deserialize error: oin is null");
		  }
		} catch (Exception e) {
		  System.out.println( "ResponseTimeTable deserialize exception: " + e.getMessage() );
		}
        }
	*/

	cTable = new ResponseTimeTable();	// Mehmet: added this instead of deserialize code above
	cTable.setNoOfSegments(noOfSegments);
        created = true;
	updated = false;
        self    = cTable;
        return cTable;
    }

   /*
    *=========================================================================
    * serialize response time table to the disk
    *=========================================================================
    */
/*
    public void serialize()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;
        
        try {            
	    if (this.size() > 0) {
              out  = new FileOutputStream( serFileName );
	      if (out != null)
                oout = new ObjectOutputStream( out );
	    }
	    if (oout != null) {
              oout.writeObject( this );
              oout.flush();
              oout.close();
	    }
        } catch( IOException e ){
            System.out.println( "ResponseTimeTable serialize exception: " + e.getMessage() );
		return;
        }
	  updated = false;
    }
*/

   /*
    *=========================================================================
    * set the number of segments variable value (noOfSegments).
    *	This function should be called after the segments are created.
    *=========================================================================
    */
    public void setNoOfSegments(int num)
    {
	  noOfSegments = num;
    }

   /*
    *=========================================================================
    * close response time table by serializing it on the disk
    *=========================================================================
    */
    public void close()
    {
	  //if (updated) serialize();
	  closed = true;
    }
    
   /*
    *=========================================================================
    * serialize on the disk if not done
    *=========================================================================
    */
    protected void finalize()
    {
        if( !closed ){
            close();
        }
    }

   /*
    *=========================================================================
    * add a new item to the response time table
    *=========================================================================
    */
    public synchronized Object put(Object key, Object value)
    {
	Object tmp = super.put(key, value);
	updated = true;
	System.out.println("ResponseTime put: " + key);
	return tmp;
    }

   /*
    *=========================================================================
    * add a new item to the response time table
    *=========================================================================
    */
    public synchronized Object put(String fname, int seg, long rtime)
    {
	ResponseTimeTableItem item = null;
	if (containsKey(fname)) {
	  item = (ResponseTimeTableItem) get(fname);
	}
	else {
	  item = new ResponseTimeTableItem(noOfSegments);
	}
	if (item == null) {
	  System.out.println("Error: cannot create a new ResponseTimeTableItem");
	  return null;
	}
	item.insertTime(seg, rtime);
	return put(fname, item);
    }

    public boolean updated()
    {
	return updated;
    }

    // getMostActiveSegment() : returns the ID number for the most active
    //	segment for the given resource
    public Integer getMostActiveSegment(String fname) {
	if (containsKey(fname)) {
	  ResponseTimeTableItem item = (ResponseTimeTableItem) get(fname);
	  return item.getMostActiveSegment();
	}
	else return null;
    }

    // getMostActiveSegmentPercentage() : tells what percentage of the total
    //	response time is coming from the most active segment
    public double getMostActiveSegmentPercentage(String fname) {
	ResponseTimeTableItem item = null;
	if (containsKey(fname)) item = (ResponseTimeTableItem) get(fname);
	else return 0;
	Integer segment = item.getMostActiveSegment();
	if (segment == null) return 0;
	int seg = segment.intValue();
	return item.getLoadPercentage(seg);
    }

    // getRequestCount() : tells the total number of request measured for the given
    //	resource since the last clear() operation
    public long getRequestCount(String fname) {
	ResponseTimeTableItem item = null;
	if (containsKey(fname)) item = (ResponseTimeTableItem) get(fname);
	else return 0;
	return item.getRequestCount();
    }

    // clear() : clears all response times recorded for the given resource
    public void clear(String fname) {
	if (containsKey(fname)) {
	  ResponseTimeTableItem item = (ResponseTimeTableItem) get(fname);
	  item.clear();
	  put(fname, item);
	}
    }

    // clear() : clears all response times recorded for all resources
    public void clear() {
	String fname = null;
	ResponseTimeTableItem item = null;

	Enumeration enum = keys();
	while( enum.hasMoreElements() ) {
	  fname = (String) enum.nextElement();
	  item = (ResponseTimeTableItem) get(fname);
	  item.clear();
	  put(fname, item);
	}
    }

}	// class ResponseTimeTable ends here


// ========================================================================


class ResponseTimeTableItem {

    public long segmentTimes[];		// array that keeps resp times for all segments (server groups) for a specific resource
    public long totalTime;		// total response time for a specific resource
    public long requestCount;		// counts the total no of requests for a specific resource
    protected int noOfSegments;

    public ResponseTimeTableItem() {}

    public ResponseTimeTableItem(int size) {
	noOfSegments = size;
	segmentTimes = new long[noOfSegments];
	int i;
	for (i=0; i<noOfSegments; i++) segmentTimes[i] = 0;
	totalTime = 0;
	requestCount = 0;
    }

    public void insertTime(int seg, long rtime) {
	if (seg >= noOfSegments) {
	  System.out.println("Error: segment no is larger than noOfSegments");
	  return;
	}
	segmentTimes[seg] += rtime;
	totalTime += rtime;
	requestCount++;
	Webpp.totalRequestCount++;
    }

    public long getTime(int seg) {
	return segmentTimes[seg];
    }

    // clear() : clears all response times recorded for this resource
    public void clear() {
	int i;
	for (i=0; i<noOfSegments; i++) segmentTimes[i] = 0;
	totalTime = 0;
	Webpp.totalRequestCount -= requestCount;
	requestCount = 0;
    }

    // getMostActiveSegment() : returns the ID number for the most active
    //	segment for a specific resource
    public Integer getMostActiveSegment() {
	int i, result = -1;
	long maxRT = 0;
	for (i=0; i<noOfSegments; i++)
	  if (segmentTimes[i] > maxRT) {
	    maxRT = segmentTimes[i];
	    result = i;
	  }
	if (result != -1) return (new Integer(result));
	else return null;
    }

    // getLoadPercentage() : tells what percentage of total response time
    //	is consumed by clients in the given segment
    public double getLoadPercentage(int seg) {
	if (noOfSegments == 0) return 0;
	return (segmentTimes[seg] * 1.0) / noOfSegments;
    }

    // getRequestCount() : tells the total number of request measured for a specific
    //	resource since the last clear() operation
    public long getRequestCount() {
	return requestCount;
    }

}	// class ResponseTimeTableItem ends here
