package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;


///////////////////////////////////////////////////////////////////////////////
//
// Mehmet: Serializable Hash Table
//
///////////////////////////////////////////////////////////////////////////////
class SerHashtable extends Hashtable implements Serializable
{
    /* class constants */
    protected static final int READ_FROM_DISK = 1;
    protected static final int NEWLY_CREATED = 2;

    /* class variables */
    protected static boolean created = false;
    protected static boolean updated = false;
    protected static String serFileName = null, fname = null;
    protected static File cTableFd;
    protected static String cTableDirName;    
    protected static SerHashtable self = null;
    protected static int status = -1;
    
    /* instance variables */
    protected boolean closed = false;
    
   /*
    *=========================================================================
    * Constructor
    *=========================================================================
    */    
    protected SerHashtable(String fname)
    {
        super();
        this.fname = fname;
    }

    protected SerHashtable()
    {
        this("serhashtable.ser");
    }
    
   /*
    *=========================================================================
    * Replication Directory factory
    *=========================================================================
    */

    public static SerHashtable factory()
    {
	return factory("serhashtable.ser");
    }

    public static SerHashtable factory(String filename)
    {
        File fd = null;
        SerHashtable cTable = null;
        boolean status;
        String line = null;
        String thekey;
        String thevalue;
        int pos;
        
        if( created ){
            // serializable hash table is singleton 
            // Assert.assert( false );
            System.out.println( Webpp.webppRootDir +
                                ": detected another instance of SerHashtable" );                                
        }
         
        cTableDirName = Webpp.webppRootDir + 
                      System.getProperty( "file.separator" ) +
                      "replication";
        cTableFd = new File( cTableDirName );
      
        if( !cTableFd.exists() ){
            status = cTableFd.mkdirs();
            Assert.assert( status == true, 
                           "SerHashtable.factory: could not create " + 
                           cTableDirName );
        }
      
        serFileName = cTableDirName + System.getProperty( "file.separator" ) + filename;
        fd = new File( serFileName );

        if( !fd.exists() ){
	  // create a new serializable hash table
	  cTable = new SerHashtable(filename);
	  cTable.setStatus(NEWLY_CREATED);
		
        } else {
          // deserialize an existing serializable hash table
	  try {
              FileInputStream in = new FileInputStream( fd );
              ObjectInputStream oin = null;
	      if (in != null) {
		oin = new ObjectInputStream( in );
	      }
	      if (oin != null) {
                cTable = (SerHashtable) oin.readObject();
		oin.close();
	      }
	      else {
		System.out.println( "SerHashtable deserialize error: oin is null");
	      }
	  } catch (Exception e) {
	    System.out.println( "SerHashtable deserialize exception: " + e.getMessage() );
	  }
	  cTable.setFilename(filename);
	  cTable.setStatus(READ_FROM_DISK);
        }
                
        created = true;
	updated = false;
        self    = cTable;
        return cTable;
    }

   /*
    *=========================================================================
    * serialize SerHashtable to the disk
    *=========================================================================
    */
    public void serialize()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;

	if (!updated) return;
        
        try {            
	    if (this.size() > 0) {
              out  = new FileOutputStream( serFileName );
	      if (out != null)
                oout = new ObjectOutputStream( out );
	    }
	    if (oout != null) {
              oout.writeObject( this );
              oout.flush();
              oout.close();
	    }
        } catch( IOException e ){
            System.out.println( "SerHashtable serialize exception: " + e.getMessage() );
		return;
        }
	updated = false;
    }


   /*
    *=========================================================================
    * set the filename for the serialization/deserialization of the hash table
    *=========================================================================
    */
    public void setFilename(String fname)
    {
	this.fname = fname;
    }


   /*
    *=========================================================================
    * isNew() : returns true if this SerHashtable object was created as a new
    *	one, i.e. not serialized from disk
    *=========================================================================
    */
    public boolean isNew()
    {
	return (status == NEWLY_CREATED);
    }


   /*
    *=========================================================================
    * setStatus() : sets the status variable of this SerHashtable object
    *=========================================================================
    */
    public void setStatus(int stat)
    {
	status = stat;
    }


   /*
    *=========================================================================
    * close SerHashtable table by serializing it on the disk
    *=========================================================================
    */
    public void close()
    {
	  if (updated) serialize();
	  closed = true;
    }
    
   /*
    *=========================================================================
    * serialize on the disk if not done
    *=========================================================================
    */
    protected void finalize()
    {
        if( !closed ){
            close();
        }
    }

   /*
    *=========================================================================
    * add a new item to the SerHashtable table
    *=========================================================================
    */
    public synchronized Object put(Object key, Object value)
    {
	Object tmp = super.put(key, value);
	updated = true;
	//System.out.println("SerHashtable put: " + key + " " + value);
	return tmp;
    }

    public boolean updated()
    {
	return updated;
    }
}   
